/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gwf_lib.h>
#include "gwf_projdata_elem.h"

typedef struct _GwfElementForeach	GwfElementForeach;
struct _GwfElementForeach
{
  GwfElementForeachFunc	func;
  gpointer		user_data;
};

typedef struct _GwfElementData		GwfElementData;
struct _GwfElementData
{
  GList		*list;
};

enum {
   ELEMENT_ADD,
   ELEMENT_CHANGE,
   ELEMENT_REMOVE,

   LAST_SIGNAL
};

static void		gwf_project_element_struct_free	(gpointer data);
static gint		gwf_element_cmp 		(gconstpointer a,gconstpointer b);
static void		gwf_element_list_free		(GList **list);
static element*		gwf_project_get_element		(GwfProject *project,gchar *name);
static GList*		gwf_project_get_element_list 	(GwfProject *project,gchar *name);
static void		foreach_element 		(gpointer data,gpointer user_data);
static GwfElementData*	gwf_project_element_get_struct	(GwfProject *project);

static guint project_signals[LAST_SIGNAL] = { 0 };

GwfProjData*
gwf_project_element_new	(GwfProject *project)
{
  GwfProjData		*tmp;
  GwfElementData	*data;
  static gint		add_signals = 1;

  g_return_val_if_fail(project != NULL,NULL);

  tmp = gwf_proj_data_new("Elements");

  data = g_malloc(sizeof(GwfElementData));
  data->list = NULL;
  gwf_proj_data_set_data(tmp,(gpointer)data);
  gwf_proj_data_set_free_func(tmp,gwf_project_element_struct_free);

  if (add_signals)
  {
    project_signals[ELEMENT_ADD] = gwf_project_new_signal(project,"element_added",gtk_marshal_NONE__STRING,GTK_TYPE_NONE,1,GTK_TYPE_STRING);
    if (!project_signals[ELEMENT_ADD])
    {
      gwf_proj_data_free(tmp);                                           ;
      return NULL;
    }

    project_signals[ELEMENT_CHANGE] = gwf_project_new_signal(project,"element_changed",gtk_marshal_NONE__STRING,GTK_TYPE_NONE,1,GTK_TYPE_STRING);
    if (!project_signals[ELEMENT_CHANGE])
    {
      gwf_proj_data_free(tmp);
      return NULL;
    }

    project_signals[ELEMENT_REMOVE] = gwf_project_new_signal(project,"element_removed",gtk_marshal_NONE__STRING,GTK_TYPE_NONE,1,GTK_TYPE_STRING);
    if (!project_signals[ELEMENT_REMOVE])
    {
      gwf_proj_data_free(tmp);
      return NULL;
    }
    add_signals = 0;
  }

  return tmp;
}

static void
gwf_project_element_struct_free (gpointer data)
{
  gwf_element_list_free(&((GwfElementData*)data)->list);
  g_free(data);
}

/*********************/
/* Element functions */
/*********************/

void
gwf_project_add_element	(GwfProject *project,element *elem)
{
  element		*tmp_elem;
  GwfElementData	*data;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(elem != NULL);

  data = gwf_project_element_get_struct(project);
  g_return_if_fail(data != NULL);

  if (gwf_project_get_element_list(project,elem->name))
    return;
  
  tmp_elem = g_malloc(sizeof(element));
  tmp_elem->name = g_strdup(elem->name);
  tmp_elem->type = g_strdup(elem->type);
  tmp_elem->value = g_strdup(elem->value);
  tmp_elem->set_func = elem->set_func;
  tmp_elem->get_func = elem->get_func;
  tmp_elem->alloc_set_func = elem->alloc_set_func;
  tmp_elem->alloc_get_func = elem->alloc_get_func;
  tmp_elem->set_code = g_strdup(elem->set_code);
  tmp_elem->get_code = g_strdup(elem->get_code);

  data->list = g_list_append(data->list,(gpointer)tmp_elem);
  
  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_ADD],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_new_element (GwfProject *project,gchar *name)
{
  element	tmp_elem;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  
  tmp_elem.name = name;
  tmp_elem.type = NULL;
  tmp_elem.value = NULL;
  tmp_elem.set_func = FALSE;
  tmp_elem.get_func = FALSE;
  tmp_elem.alloc_set_func = FALSE;
  tmp_elem.alloc_get_func = FALSE;
  tmp_elem.set_code = NULL;
  tmp_elem.get_code = NULL;
  
  gwf_project_add_element(project,&tmp_elem);
}

void
gwf_project_del_element	(GwfProject *project,gchar *name)
{
  GList			*tmp_list;
  GwfElementData	*data;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  
  data = gwf_project_element_get_struct(project);
  g_return_if_fail(data != NULL);

  tmp_list = gwf_project_get_element_list(project,name);
  if (!tmp_list)
    return;
    
  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_REMOVE],name);

  g_free(((element*)tmp_list->data)->name);
  g_free(((element*)tmp_list->data)->type);
  g_free(((element*)tmp_list->data)->value);
  g_free(((element*)tmp_list->data)->set_code);
  g_free(((element*)tmp_list->data)->get_code);
  g_free(tmp_list->data);

  data->list = g_list_remove(data->list,tmp_list->data);

  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_name (GwfProject *project,gchar *name,gchar *newname)
{
  element		*tmp_elem;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  g_return_if_fail(newname != NULL);
  
  if (gwf_project_get_element_list(project,newname))
    return;
    
  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;
    
  g_free(tmp_elem->name);
  tmp_elem->name = g_strdup(newname);

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_type (GwfProject *project,gchar *name,gchar *type)
{
  element		*tmp_elem;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  
  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;
    
  g_free(tmp_elem->type);
  tmp_elem->type = g_strdup(type);

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_value	(GwfProject *project,gchar *name,gchar *value)
{
  element		*tmp_elem;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  
  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;
    
  g_free(tmp_elem->value);
  tmp_elem->value = g_strdup(value);

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_setfunc	(GwfProject *project,gchar *name,gboolean setfunc,gboolean alloc)
{
  element		*tmp_elem;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  
  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;
    
  tmp_elem->set_func = setfunc;
  tmp_elem->alloc_set_func = alloc;

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_setcode(GwfProject *project,gchar *name,gchar *code)
{
  element		*tmp_elem;

  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);

  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;

  g_free(tmp_elem->set_code);
  tmp_elem->set_code = g_strdup(code);

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_getfunc (GwfProject *project,gchar *name,gboolean getfunc,gboolean alloc)
{
  element		*tmp_elem;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);
  
  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;
    
  tmp_elem->get_func = getfunc;
  tmp_elem->alloc_get_func = alloc;

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

void
gwf_project_element_set_getcode(GwfProject *project,gchar *name,gchar *code)
{
  element		*tmp_elem;

  g_return_if_fail(project != NULL);
  g_return_if_fail(name != NULL);

  tmp_elem = gwf_project_get_element(project,name);
  if (!tmp_elem)
    return;

  g_free(tmp_elem->get_code);
  tmp_elem->get_code = g_strdup(code);

  gtk_signal_emit(GTK_OBJECT(project),project_signals[ELEMENT_CHANGE],tmp_elem->name);
  gwf_project_set_changed(project,TRUE);
}

gchar*
gwf_project_element_get_type (GwfProject *project,gchar *name)
{
  element		*tmp_elem;
  
  g_return_val_if_fail(project != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);
  
  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->type : NULL;
}

gchar*
gwf_project_element_get_value (GwfProject *project,gchar *name)
{
  element		*tmp_elem;
  
  g_return_val_if_fail(project != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);
  
  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->value : NULL;
}

gboolean
gwf_project_element_get_setfunc	(GwfProject *project,gchar *name)
{
  element		*tmp_elem;
  
  g_return_val_if_fail(project != NULL,FALSE);
  g_return_val_if_fail(name != NULL,FALSE);
  
  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->set_func : FALSE;
}

gboolean
gwf_project_element_get_setalloc(GwfProject *project,gchar *name)
{
  element		*tmp_elem;
  
  g_return_val_if_fail(project != NULL,FALSE);
  g_return_val_if_fail(name != NULL,FALSE);
  
  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->alloc_set_func : FALSE;
}

gchar*
gwf_project_element_get_setcode (GwfProject *project,gchar *name)
{
  element		*tmp_elem;

  g_return_val_if_fail(project != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);

  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->set_code : NULL;
}

gboolean
gwf_project_element_get_getfunc	(GwfProject *project,gchar *name)
{
  element		*tmp_elem;
  
  g_return_val_if_fail(project != NULL,FALSE);
  g_return_val_if_fail(name != NULL,FALSE);
  
  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->get_func : FALSE;
}

gboolean
gwf_project_element_get_getalloc(GwfProject *project,gchar *name)
{
  element		*tmp_elem;
  
  g_return_val_if_fail(project != NULL,FALSE);
  g_return_val_if_fail(name != NULL,FALSE);
  
  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->alloc_get_func : FALSE;
}

gchar*
gwf_project_element_get_getcode (GwfProject *project,gchar *name)
{
  element		*tmp_elem;

  g_return_val_if_fail(project != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);

  tmp_elem = gwf_project_get_element(project,name);
  return (tmp_elem) ? tmp_elem->get_code : NULL;
}

void
gwf_project_element_foreach (GwfProject *project,GwfElementForeachFunc func,gpointer user_data)
{
  GwfElementForeach	tmp;
  GwfElementData	*data;
  
  g_return_if_fail(project != NULL);
  g_return_if_fail(func != NULL);
  
  tmp.func = func;
  tmp.user_data = user_data;

  data = gwf_project_element_get_struct(project);
  g_return_if_fail(data != NULL);

  g_list_foreach(data->list,foreach_element,(gpointer)&tmp);
}

element*
gwf_project_element_find (GwfProject *project,gchar *name)
{
  g_return_val_if_fail(project != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);

  return gwf_project_get_element(project,name);
}

/*****************************/
/* Internal search functions */
/*****************************/

static element*	
gwf_project_get_element	(GwfProject *project,gchar *name)
{
  GList		*tmp_list;
  g_return_val_if_fail(project != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);
  
  tmp_list = gwf_project_get_element_list(project,name);
  return (tmp_list) ? (element*)(tmp_list->data) : NULL;
}

static GList*
gwf_project_get_element_list (GwfProject *project,gchar *name)
{
  GwfElementData	*data;

  data = gwf_project_element_get_struct(project);
  g_return_val_if_fail(data != NULL,NULL);

  return g_list_find_custom(data->list, (gpointer)name, (GCompareFunc)gwf_element_cmp);
}

static GwfElementData*
gwf_project_element_get_struct (GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);

  return (GwfElementData*)gwf_proj_data_get_data(gwf_data_list_get(gwf_project_get_data_list(project),"Elements"));
}
/************************/
/* GList free functions */
/************************/

static void
gwf_element_list_free (GList **list)
{
  GList		*tmp_list;

  for ( tmp_list = *list ; tmp_list ; tmp_list = g_list_next(tmp_list) )
  {
    g_free( ((element*)tmp_list->data)->type);
    g_free( ((element*)tmp_list->data)->name);
    g_free( ((element*)tmp_list->data)->value);
    g_free( ((element*)tmp_list->data)->set_code);
    g_free( ((element*)tmp_list->data)->get_code);

    g_free( tmp_list->data);
  }
  g_list_free(*list);
  *list = NULL;
}

/***************************/
/* GList compare functions */
/***************************/

static gint
gwf_element_cmp (gconstpointer a,gconstpointer b)
{
  return strcmp( ((element*)a)->name, (gchar*) b);
}

/*********************/
/* Foreach functions */
/*********************/

static void
foreach_element (gpointer data,gpointer user_data)
{
  GwfElementForeach	*tmp = user_data;
  
  tmp->func((element*)data,tmp->user_data);
}


